"""Create a mindy index for a database

See the usage for more information.
"""

import sys
from xml.sax import handler
from bsddb3 import db, dbshelve
import Martel

__version__ = "0.1"

def usage(outfile):
    outfile.write("""\
mindy_index %(version)s
Usage:
  %(argv0)s [opts] filename1 filename2 ...
where opts can be:
 --dbhome <dirname>       The control directory (defaults to '.mindy_dbhome')
 --dbname <filename>      Name of the database (required)
 --format <name>          The name of the Martel format expression (required)
                              (eg, Martel.formats.swissprot38)
 --record-tag <name>      The element name containing a record (required)
                              (eg, swissprot38)
 --progress <count>       Print progress information to stderr after every
                              'count' records.  Defaults to 0 meaning no info.

The indexing can use one of three styles, one of which must be specified.
 Simple indexing:
  --identifier <tagname>  The identifier is found inside the given tag.
                            Only one identifier is allowed. (required)
  --alias <tagname>       The alias is found inside the given tag.
                            Multiple aliases are allowed. (optional)
  --keyword <tagname>     The keyword is found inside the given tag.
                            Multiple keywords are allowed. (optional)
 XSLT indexing:
  --xslt <filename>       An XSLT transformation producing an XML file
                            with <identifier>, <alias> and <keyword> elements.
                            (not yet implemented)
 Python indexing:
  --function <name>       The name of a Python function to import which
                            follows the Indexer protocol
                            (not yet implemented)

 --help                   Print this usage information
 --version                Print the version number

Create a mindy index for one or more data files.  The index is usable
by programs like "mindy_search."

Example:
%% %(argv0)s --format Martel.formats.swissprot38.format \\
  --record-tag swissprot38_record --dbname swiss \\
  --identifier entry_name --alias ac_number --progress 100\\
  /home/dalke/ftps/swissprot/sprot38.dat
%% mindy_search --dbname swiss --identifier 100K_RAT
ID   100K_RAT       STANDARD;      PRT;   889 AA.
AC   Q62671;
DT   01-NOV-1997 (Rel. 35, Created)
 ...

""" % {"argv0": sys.argv[0],
       "version": __version__}
                  )

def version(outfile):
    outfile.write("mindy_index %s\n" % __version__)

class SimpleIndexer(handler.ContentHandler):
    def __init__(self, mindy, record_tag, identifier, aliases, keywords):
        self.mindy = mindy
        self.record_tag = record_tag
        
        fields = {}
        fields[identifier] = 1
        for k in aliases:
            fields[k] = 1
        for k in keywords:
            fields[k] = 1
        self.fields = fields.keys()

        self._wanted_elements = self.fields + [record_tag]
        
        self.identifier = identifier
        self.aliases = aliases
        self.keywords = keywords

    def startDocument(self):
        self.pos = 0
        self.capture = 0

    def startElement(self, name, attrs):
        if name in self.fields:
            self.capture = 1
            self.s = ""
        if name == self.record_tag:
            self.results = {"identifier": None,
                            "aliases": [],
                            "keywords": [],
                            }
            self.startpos = self.pos
            
    def characters(self, s):
        if self.capture:
            self.s = self.s + s
        self.pos = self.pos + len(s)
        
    def endElement(self, name):
        if name in self.fields:
            self.capture = 0
            if name == self.identifier:
                self.results["identifier"] = self.s
            if name in self.aliases:
                self.results["aliases"].append(self.s)
            if name in self.keywords:
                self.results["keywords"].append(self.s)

        if name == self.record_tag:
            self.mindy.add_results(startpos = self.startpos,
                                   endpos = self.pos,
                                   **self.results)
        

# mindy_data:
#   filenames -> dict mapping 'abbrev' identifier to filename
#   abbrev -> current max identifer
#   format -> string used to get the format name
#   record -> tag name used to define a record
# identifiers:
#   id -> (abbrev, startpos, endpos)
# aliases:
#   alias -> id
# keywords:
#   keyword -> id

class Mindy:
    def __init__(self, mindy_data, identifiers, aliases, keywords):
        self.mindy_data = mindy_data
        self.identifiers = identifiers
        self.aliases = aliases
        self.keywords = keywords
        self.filename = None
        self._filenames = {}
        self._abbrevs = {}

    def add_filename(self, filename):
        assert not self._abbrevs.has_key(filename), "duplicate filename"
        
        abbrev = self.mindy_data["abbrev"]
        self.mindy_data["abbrev"] = abbrev + 1
        
        self._filenames[str(abbrev)] = filename
        self._abbrevs[filename] = str(abbrev)

        self.mindy_data["filenames"] = self._filenames

    def use_filename(self, filename):
        if not self._abbrevs.has_key(filename):
            self.add_filename(filename)
        
        self.filename = filename
        self.abbrev = self._abbrevs[filename]

    def add_results(self, startpos, endpos, identifier, aliases, keywords):
        #print "Adding", identifier, startpos, endpos
        self.identifiers[identifier] = (self.abbrev, startpos, endpos)
        for alias in aliases:
            try:
                self.aliases[alias] = self.aliases[alias] + (identifier,)
            except KeyError:
                self.aliases[alias] = (identifier,)
        for keyword in keywords:
            try:
                self.keywords[alias] = self.keywords[alias] + (identifier,)
            except KeyError:
                self.keywords[alias] = (identifier,)

class Progress:
    def __init__(self, mindy, count):
        self.__mindy = mindy
        self.__count = count
        self.__n = 0
    def __getattr__(self, name):
        return getattr(self.__mindy, name)
    def add_results(self, startpos, endpos, identifier, aliases, keywords):
        self.__mindy.add_results(startpos, endpos, identifier, aliases,
                                 keywords)
        self.__n = self.__n + 1
        if self.__n % self.__count == 0:
            print >>sys.stderr, "Added record %s: %s" % (self.__n, identifier)

def _init(mindy):
    mindy.mindy_data["filenames"] = {}
    mindy.mindy_data["abbrev"] = 0

def create(dbhome, dbname):
    env = db.DBEnv()
    envflags = db.DB_THREAD | db.DB_INIT_MPOOL | db.DB_INIT_LOCK
    try:
        env.open(dbhome, envflags | db.DB_CREATE)
    except db.DBNoSuchFileError, err:
        errno, msg = err
        raise IOError(errno, msg, dbhome)

    mindy_data = dbshelve.DBShelf(env)
    mindy_data.open(dbname, "mindy_data", db.DB_HASH,
                    db.DB_CREATE | db.DB_TRUNCATE, 0644)

    identifiers = dbshelve.DBShelf(env)
    identifiers.open(dbname, "identifiers", db.DB_HASH, db.DB_CREATE, 0644)

    aliases = dbshelve.DBShelf(env)
    aliases.open(dbname, "aliases", db.DB_HASH, db.DB_CREATE, 0644)

    keywords = dbshelve.DBShelf(env)
    keywords.open(dbname, "keywords", db.DB_HASH, db.DB_CREATE, 0644)

    mindy = Mindy(mindy_data, identifiers, aliases, keywords) 
    _init(mindy)

    return mindy

def load_format(format_name):
    fields = format_name.split(".")
    obj = __import__(fields[0])
    name = fields[0]
    try:
        for field in fields[1:]:
            name = name + "." + field
            __import__(name)
    except ImportError:
        pass
    for field in fields[1:]:
        obj = getattr(obj, field)
    return obj

def main(argv):
    import getopt
    try:
        opts, filenames = getopt.getopt(argv[1:],
                                        "",
                                        ["dbhome=", "dbname=",
                                         "format=", "record-tag=",
                                         "progress=",
                                         "identifier=", "alias=", "keyword=",
                                         "xslt=",
                                         "function=",
                                         "help", "version",
                                         ])
    except getopt.GetoptError, msg:
        print >>sys.stderr, str(msg)
        return 1

    format = None
    record_tag = None
    progress = 0
    dbhome = ".mindy_dbhome"
    dbname = None
    identifier = None
    aliases = []
    keywords = []
    xslt = None
    function = None

    for k, v in opts:
        if k == "--format":
            format = v
        elif k == "--record-tag":
            record_tag = v
        elif k == "--progress":
            progress = int(v)
        elif k == "--dbhome":
            dbhome = v
        elif k == "--dbname":
            dbname = v
        elif k == "--identifier":
            identifier = v
        elif k == "--alias":
            aliases.append(v)
        elif k == "--keywords":
            keywords.append(v)
        elif k == "--xslt":
            xslt = v
        elif k == "--function":
            function = v
        elif k == "--help":
            usage(sys.stdout)
            return 0
        elif k == "--version":
            version(sys.stdout)
            return 0
        else:
            raise AssertionError, "unknown argument %r" % k
            
    if format is None:
        print >>sys.stderr, "Format name must be specified with --format"
        return 1
    if record_tag is None:
        print >>sys.stderr, "Record tag name must be specified with --record"
        return 1
    if dbname is None:
        print >>sys.stderr, "Database filename must be specified with --dbname"
        return 1

    count = 0
    if identifier is not None:
        using = "--identifier"
        count = count + 1
    if xslt is not None:
        using = "--xslt"
        count = count + 1
    if function is not None:
        using = "--function"
        count = count + 1
        
    if count == 0:
        print >>sys.stderr, "Must specify one of --identifier, --xslt or --function"
        return 1

    if count != 1:
        print >>sys.stderr, "Can only specify one of --identifier, --xslt or --function"
        return 1

    if identifier is None:
        if aliases:
            print >>sys.stderr, "Cannot specify --aliases when using", using
            return 1
        if keywords:
            print >>sys.stderr, "Cannot specifiy --keywords when using", using
            return 1

    if not filenames:
        print >>sys.stderr, "No filenames specified"
        return 1

    if progress < 0:
        print >>sys.stderr, "Progress value cannot be negative - you wanted", progress
        return 1

    mindy = create(dbhome, dbname)
    if progress != 0:
        mindy = Progress(mindy, progress)

    if identifier is not None:
        indexer = SimpleIndexer(mindy, record_tag, identifier, aliases,
                                keywords)
    elif xslt is not None:
        indexer = make_xslt_indexer(mindy, xslt)
    elif function is not None:
        indexer = make_function_indexer(mindy, function)

    format = load_format(format)

    if hasattr(indexer, "_wanted_elements"):
        format = Martel.select_names(format, indexer._wanted_elements)
    
    parser = format.make_parser()
    parser.setContentHandler(indexer)
    for filename in filenames:
        mindy.use_filename(filename)
        parser.parseFile(open(filename, "rb"))
        

if __name__ == "__main__":
    retval = main(sys.argv)
    if retval:
        sys.exit(retval)
