"""Search a Mindy indexed database

See the usage for more information.
"""

import sys
from bsddb3 import db, dbshelve

__version__ = "0.1"

def usage(outfile):
    outfile.write("""\
mindy_search %(version)s
Usage:
  %(argv0)s [opts]
where opts can be:
 --dbhome <dirname>       The control directory (defaults to '.mindy_dbhome')
 --dbname <filename>      Name of the database (required)

Three search methods are available and one must be specified
 --identifier <name>      Search for the given identifier
 --alias <name>           Search for all records with the given alias
 --keyword <name>         Search for all records with the given keywords

 --show-record            If set to 1 (the default), print all matching
                          records to stdout.  If set to 0, print each
                          found identifier record on its own line.  (The
                          identifier is repr escaped excluding the quotes.)

 --help                   Print this usage information
 --version                Print the version number

Example:
%% mindy_search.py --dbname swiss --alias P32261
ID   ANT3_MOUSE     STANDARD;      PRT;   465 AA.
AC   P32261;
DT   01-OCT-1993 (Rel. 27, Created)
 ...
""" % {"argv0": sys.argv[0],
       "version": __version__,
       })

def version(outfile):
    outfile.write("mindy_search %s\n" % __version__)


class MindySearch:
    def __init__(self, mindy_data, identifiers, aliases, keywords):
        self.mindy_data = mindy_data
        self.identifiers = identifiers
        self.aliases = aliases
        self.keywords = keywords

        self._filenames = self.mindy_data["filenames"]

    def __getitem__(self, id):
        abbrev, startpos, endpos = self.identifiers[id]
        filename = self._filenames[abbrev]
        infile = open(filename, "rb")
        infile.seek(startpos)
        return infile.read(endpos - startpos)



def mindy_open(dbhome, dbname):
    env = db.DBEnv()
    envflags = db.DB_THREAD | db.DB_INIT_MPOOL | db.DB_INIT_LOCK
    env.open(dbhome, envflags)

    mindy_data = dbshelve.DBShelf(env)
    mindy_data.open(dbname, "mindy_data", db.DB_HASH, db.DB_CREATE, 0644)

    identifiers = dbshelve.DBShelf(env)
    identifiers.open(dbname, "identifiers", db.DB_HASH, db.DB_CREATE, 0644)

    aliases = dbshelve.DBShelf(env)
    aliases.open(dbname, "aliases", db.DB_HASH, db.DB_CREATE, 0644)

    keywords = dbshelve.DBShelf(env)
    keywords.open(dbname, "keywords", db.DB_HASH, db.DB_CREATE, 0644)

    return MindySearch(mindy_data, identifiers, aliases, keywords)

def main(argv):
    import getopt
    try:
        opts, filenames = getopt.getopt(argv[1:],
                                        "",
                                        ["dbhome=", "dbname=",
                                         "identifier=", "alias=", "keyword=",
                                         "show-record=",
                                         "help", "version"])
    except getopt.GetoptError, msg:
        print >>sys.stderr, str(msg)
        usage(sys.stderr)
        return 1

    dbhome = ".mindy_dbhome"
    dbname = None
    identifier = None
    alias = None
    keyword = None
    show_record = 1
    
    if filenames:
        print >>sys.stderr, "Do not understand option", repr(filenames[0])
        return 1
    for k, v in opts:
        if k == "--dbhome":
            dbhome = v
        elif k == "--dbname":
            dbname = v
        elif k == "--identifier":
            identifier = v
        elif k == "--alias":
            alias = v
        elif k == "--keyword":
            keyword = v
        elif k == "--show-record":
            if v not in ("0", "1"):
                print >>sys.stderr, "--show-record can only be 0 or 1, not", v
                return 1
            show_record = int(v)
        elif k == "--help":
            usage(sys.stdout)
            return 0
        elif k == "--version":
            version(sys.stdout)
            return 0
        else:
            raise AssertionError, "unknown argument %r" % v

    if dbname is None:
        print >>sys.stderr, "Database filename must be specified with --dbname"
        return 1
    
    count = 0
    if identifier is not None:
        count = count + 1
    elif alias is not None:
        count = count + 1
    elif keyword is not None:
        count = count + 1
    if count == 0:
        print >>sys.stderr, "Must specify one of --identifier, --alias or --keyword"
        return 1
    if count != 1:
        print >>sys.stderr, "Can only specify one of --identifier, --alias or --keyword"
        return 1

    mindy_search = mindy_open(dbhome, dbname)

    results = []
    if identifier is not None:
        if mindy_search.identifiers.has_key(identifier):
            results = [identifier]
    elif alias is not None:
        results = mindy_search.aliases.get(alias, [])
    elif keyword is not None:
        results = mindy_search.keywords.get(keyword, [])

    if show_record:
        for id in results:
            sys.stdout.write(mindy_search[id])
    else:
        for id in results:
            print repr(id)[1:-1]
    
if __name__ == "__main__":
    main(sys.argv)
